// Shine's Snake
//
//   artworks by Gundrosen
//   coding by Shine
//
// Credits:
//   based on Hello World for PSP by nem
//   Exit Game test by TyRaNiD
//   function stubs: PSPDev - Browser Api (by djhuevo & neofar, http://pspdev.ofcode.com/api.php)
//   controller function usage based on a program by skippy911
//   and many hints from people at http://forums.ps2dev.org

#include <pspkernel.h>
#include <pspdisplay.h>
#include <pspctrl.h>
#include <stdlib.h>
#include <string.h>

/* Define the module info section */
PSP_MODULE_INFO("SNAKE", 0, 1, 1);

/* Define the main thread's attribute value (optional) */
PSP_MAIN_THREAD_ATTR(THREAD_ATTR_USER | THREAD_ATTR_VFPU);

extern u32 background_start[];
extern u32 headR_start[];
extern u32 headT_start[];
extern u32 headL_start[];
extern u32 headB_start[];
extern u32 tailR_start[];
extern u32 tailT_start[];
extern u32 tailL_start[];
extern u32 tailB_start[];
extern u32 bodyLT_start[];
extern u32 bodyLB_start[];
extern u32 bodyRT_start[];
extern u32 bodyRB_start[];
extern u32 bodyLR_start[];
extern u32 bodyTB_start[];
extern u32 apple_start[];
extern u32 fly_start[];

u32* tails[] = { tailR_start, tailT_start, tailL_start, tailB_start };
u32* heads[] = { headR_start, headT_start, headL_start, headB_start };
u32* bodies[] = { bodyLR_start, bodyTB_start, bodyLR_start, bodyTB_start };

#define RIGHT 0
#define TOP 1
#define LEFT 2
#define BOTTOM 3

/* game structs and variables */
typedef struct
{
	int x;
	int y;
	int direction;
} Cell;

#define RGB(r, g, b) ((r)|((g)<<8)|((b)<<16))
#define IS_ALPHA(value) (((value)&0xff000000)==0)

#define TEXT_COLOR RGB(0, 0, 0)

Cell* cells;
Cell target;
u32* targetImage;
int dx;
int dy;
int x;
int y;
int cellStart;
int cellEnd;
int maxCells = 60*34;
int score;
int high = 0;

static unsigned seed1;    
static unsigned seed2; 

//constants
#define SCREEN_WIDTH 480
#define SCREEN_HEIGHT 272

#define	PSP_PIXEL_FORMAT 3  // 32 bit RGBA
#define	PSP_LINE_SIZE 512   // in u32

//480*272 = 60*38
#define CMAX_X 60
#define CMAX_Y 38
#define CMAX2_X 30
#define CMAX2_Y 19
#define CMAX4_X 15
#define CMAX4_Y 9

//variables
u32* pg_vramtop = (u32*)(0x04000000+0x40000000);
long pg_screenmode;
long pg_showframe;
long pg_drawframe;

void pgWaitVn(unsigned long count)
{
	for (; count>0; --count) {
		sceDisplayWaitVblankStart();
	}
}

void pgWaitV()
{
	sceDisplayWaitVblankStart();
}

u32* pgGetVramAddr(unsigned long x,unsigned long y)
{
	return pg_vramtop+(y * PSP_LINE_SIZE)+x+(pg_drawframe?PSP_LINE_SIZE*SCREEN_HEIGHT:0);
}

void pgScreenFrame(long mode,long frame)
{
	pg_screenmode=mode;
	frame=(frame?1:0);
	pg_showframe=frame;
	if (mode==0) {
		//screen off
		pg_drawframe=frame;
		sceDisplaySetFrameBuf(0,0,0,1);
	} else if (mode==1) {
		//show/draw same
		pg_drawframe=frame;
		sceDisplaySetFrameBuf((char*)(pg_vramtop+(pg_drawframe?PSP_LINE_SIZE*SCREEN_HEIGHT:0)),PSP_LINE_SIZE,PSP_PIXEL_FORMAT,1);
	} else if (mode==2) {
		//show/draw different
		pg_drawframe=(frame?0:1);
		sceDisplaySetFrameBuf((char*)(pg_vramtop+(pg_drawframe?PSP_LINE_SIZE*SCREEN_HEIGHT:0)),PSP_LINE_SIZE,PSP_PIXEL_FORMAT,1);
	}
}

void pgInit()
{
	sceDisplaySetMode(0,SCREEN_WIDTH,SCREEN_HEIGHT);
	pgScreenFrame(0,0);
}

void pgScreenFlip()
{
	pg_showframe=(pg_showframe?0:1);
	pg_drawframe=(pg_drawframe?0:1);
	sceDisplaySetFrameBuf((char*)(pg_vramtop+(pg_showframe?PSP_LINE_SIZE*SCREEN_HEIGHT:0)),PSP_LINE_SIZE,PSP_PIXEL_FORMAT,0);
}

void pgScreenFlipV()
{
	pgWaitV();
	pgScreenFlip();
}

void pgFlipDrawFrame()
{
	pg_drawframe=(pg_drawframe?0:1);
}

void pgFlipShowFrameV()
{
	pgWaitV();
	pg_showframe=(pg_showframe?0:1);
	sceDisplaySetFrameBuf((char*)(pg_vramtop+(pg_showframe?PSP_LINE_SIZE*SCREEN_HEIGHT:0)),PSP_LINE_SIZE,PSP_PIXEL_FORMAT,0);
}

/* Exit callback */
int exit_callback(void)
{
	free(cells);
	sceKernelExitGame();

	return 0;
}

/* Callback thread */
void CallbackThread(void *arg)
{
	int cbid;

	cbid = sceKernelCreateCallback("Exit Callback", exit_callback, NULL);
	sceKernelRegisterExitCallback(cbid);

	sceKernelSleepThreadCB();
}

/* Sets up the callback thread and returns its thread id */
int SetupCallbacks(void)
{
	int thid = 0;

	thid = sceKernelCreateThread("update_thread", CallbackThread, 0x11, 0xFA0, 0, 0);
	if(thid >= 0)
	{
		sceKernelStartThread(thid, 0, 0);
	}

	return thid;
}

void initSeeds() 
{ 
   seed1 = sceKernelLibcTime((void *) 0); 
   seed2 = sceKernelLibcTime((void *) 0); 
} 

unsigned int myRand() 
{ 
   seed1 = (seed1 + 46906481) ^ seed2; 
   seed2 = seed1 ^ ( ((seed2<<3) | (seed2 >> 29)) + 103995407); 
    
   return seed1 - seed2; 
}

const char digitFont[] = {
	1,1,1,1,
	1,0,0,1,
	1,0,0,1,
	1,0,0,1,
	1,1,1,1,
	
	0,0,0,1,
	0,0,0,1,
	0,0,0,1,
	0,0,0,1,
	0,0,0,1,
	
	1,1,1,1,
	0,0,0,1,
	1,1,1,1,
	1,0,0,0,
	1,1,1,1,
	
	1,1,1,1,
	0,0,0,1,
	1,1,1,1,
	0,0,0,1,
	1,1,1,1,
	
	1,0,0,1,
	1,0,0,1,
	1,1,1,1,
	0,0,0,1,
	0,0,0,1,
	
	1,1,1,1,
	1,0,0,0,
	1,1,1,1,
	0,0,0,1,
	1,1,1,1,
	
	1,1,1,1,
	1,0,0,0,
	1,1,1,1,
	1,0,0,1,
	1,1,1,1,
	
	1,1,1,1,
	0,0,0,1,
	0,0,0,1,
	0,0,0,1,
	0,0,0,1,
	
	1,1,1,1,
	1,0,0,1,
	1,1,1,1,
	1,0,0,1,
	1,1,1,1,

	1,1,1,1,
	1,0,0,1,
	1,1,1,1,
	0,0,0,1,
	1,1,1,1};

void print7Segment(int x, int y, int digit)
{
	u32* vram = pgGetVramAddr(x, y);
	int xo, yo;
	int i = digit * 5*4;
	for (yo = 0; yo < 5; yo++) {
		for (xo = 0; xo < 4; xo++) {
			if (digitFont[i]) {
				vram[xo + yo * PSP_LINE_SIZE] = TEXT_COLOR;
			} else {
				vram[xo + yo * PSP_LINE_SIZE] = background_start[x + xo + (y + yo) * SCREEN_WIDTH];
			}
			i++;
		}
	}
		
}

void printDecimal(int x, int y, int color, int value)
{
	int i;
	int digits[6];
	for (i = 5; i >= 0; i--) {
		digits[i] = value%10;
		value /= 10;
	}
	i = 0;
	while (i < 5 && digits[i] == 0) digits[i++] = -1;
	for (i = 0; i < 6; i++) {
		if (digits[i] >= 0) {
			print7Segment(x, y, digits[i]);
			x += 6;
		}
	}
}

void fillCell(int x, int y)
{
	u32* vram = pgGetVramAddr(x * 16, y * 16);
	int xo, yo;
	for (yo = 0; yo < 16; yo++) {
		for (xo = 0; xo < 16; xo++) {
			vram[xo + yo * PSP_LINE_SIZE] = background_start[16*x + xo + (16*y + yo) * SCREEN_WIDTH];
		}
	}
}

void plotCellImage(int x, int y, u32* image)
{
	u32* vram = pgGetVramAddr(x * 16, y * 16);
	int xo, yo;
	for (yo = 0; yo < 16; yo++) {
		for (xo = 0; xo < 16; xo++) {
			if (!IS_ALPHA(*image)) {
				vram[xo + yo * PSP_LINE_SIZE] = *image;
			} else vram[xo + yo * PSP_LINE_SIZE] = background_start[16*x + xo + (16*y + yo) * SCREEN_WIDTH];
			image++;
		}
	}
}

void createRandomTarget()
{
	target.x = (myRand() % 21) + 1;
	target.y = (myRand() % 15) + 1;
	if (myRand() % 2) {
		targetImage = apple_start;
	} else {
		targetImage = fly_start;
	}
}

void writeByte(int fd, unsigned char data)
{
	sceIoWrite(fd, &data, 1);
}


/*void screenshot()
{
	const char tgaHeader[] = {0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	const int width = SCREEN_WIDTH;
	const int lineWidth = PSP_LINE_SIZE;
	const int height = SCREEN_HEIGHT;
	unsigned char lineBuffer[width*4];
	u32* vram = pgGetVramAddr(0, 0);
	int x, y;
	int fd = sceIoOpen("ms0:/snake-screenshot.tga", PSP_O_CREAT | PSP_O_TRUNC | PSP_O_WRONLY, 0777);
	if (!fd) return;
	sceIoWrite(fd, tgaHeader, sizeof(tgaHeader));
	writeByte(fd, width & 0xff);
	writeByte(fd, width >> 8);
	writeByte(fd, height & 0xff);
	writeByte(fd, height >> 8);
	writeByte(fd, 24);
	writeByte(fd, 0);
	for (y = height - 1; y >= 0; y--) {
		for (x = 0; x < width; x++) {
			u32 color = vram[y * lineWidth + x];
			unsigned char red = color & 0xff;
			unsigned char green = (color >> 8) & 0xff;
			unsigned char blue = (color >> 16) & 0xff;
			lineBuffer[3*x] = blue;
			lineBuffer[3*x+1] = green;
			lineBuffer[3*x+2] = red;
		}
		sceIoWrite(fd, lineBuffer, width * 3);
	}
	sceIoClose(fd);
}*/
	
void newGame()
{
	pgFlipDrawFrame();
	u32* vram = pgGetVramAddr(0, 0);
	int xo, yo;
	u32* image = background_start;
	for (yo = 0; yo < SCREEN_HEIGHT; yo++) {
		for (xo = 0; xo < SCREEN_WIDTH; xo++) {
			if (!IS_ALPHA(*image)) vram[xo + yo * PSP_LINE_SIZE] = (*image)& 0xffffff;
			image++;
		}
	}
	
	dx = 0;
	dy = 0;
	x = 12;
	y = 8;
	cellStart = 0;
	cellEnd = 1;
	cells[cellStart].x = x;
	cells[cellStart].y = y;
	cells[cellStart].direction = RIGHT;
	cells[cellEnd].x = ++x;
	cells[cellEnd].y = y;
	cells[cellEnd].direction = RIGHT;
	createRandomTarget();
	plotCellImage(cells[cellStart].x, cells[cellStart].y, tailR_start);
	plotCellImage(cells[cellEnd].x, cells[cellEnd].y, headR_start);

	score = 0;

	pgFlipShowFrameV();
}

void keyboardControl()
{
	SceCtrlData pad;
	sceDisplayWaitVblankStart();
	sceCtrlReadBufferPositive(&pad, 1); 
	if (pad.Buttons & PSP_CTRL_UP) { 
		dx = 0;
		dy = -1;
	} else if (pad.Buttons & PSP_CTRL_DOWN) { 
		dx = 0;
		dy = 1;
	} else if (pad.Buttons & PSP_CTRL_LEFT) { 
		dx = -1;
		dy = 0;
	} else if (pad.Buttons & PSP_CTRL_RIGHT) { 
		dx = 1;
		dy = 0;
	} /*else if (pad.Buttons & PSP_CTRL_CIRCLE) {
		screenshot();
	}*/
}

void move()
{
	if (dx == 0 && dy == 0) return;
	int lastDirection = cells[cellEnd].direction;
	int lastX = cells[cellEnd].x;
	int lastY = cells[cellEnd].y;
	x += dx;
	y += dy;
	cellEnd++;
	if (cellEnd == maxCells) cellEnd = 0;
	cells[cellEnd].x = x;
	cells[cellEnd].y = y;
	if (dx == 1) {
		cells[cellEnd].direction = RIGHT;
	} else if (dy == -1) {
		cells[cellEnd].direction = TOP;
	} else if (dx == -1) {
		cells[cellEnd].direction = LEFT;
	} else if (dy == 1) {
		cells[cellEnd].direction = BOTTOM;
	}
	int direction = cells[cellEnd].direction;
	if (lastDirection == RIGHT && direction == TOP) {
		plotCellImage(lastX, lastY, bodyLT_start);
	} else if (lastDirection == TOP && direction == LEFT) {
		plotCellImage(lastX, lastY, bodyLB_start);
	} else if (lastDirection == LEFT && direction == BOTTOM) {
		plotCellImage(lastX, lastY, bodyRB_start);
	} else if (lastDirection == BOTTOM && direction == RIGHT) {
		plotCellImage(lastX, lastY, bodyRT_start);
	} else if (lastDirection == RIGHT && direction == BOTTOM) {
		plotCellImage(lastX, lastY, bodyLB_start);
	} else if (lastDirection == TOP && direction == RIGHT) {
		plotCellImage(lastX, lastY, bodyRB_start);
	} else if (lastDirection == LEFT && direction == TOP) {
		plotCellImage(lastX, lastY, bodyRT_start);
	} else if (lastDirection == BOTTOM && direction == LEFT) {
		plotCellImage(lastX, lastY, bodyLT_start);
	} else {
		plotCellImage(lastX, lastY, bodies[direction]);
	}
	plotCellImage(cells[cellEnd].x, cells[cellEnd].y, heads[cells[cellEnd].direction]);
	if (x == target.x && y == target.y) {
		createRandomTarget();
		score++;
	} else {
		fillCell(cells[cellStart].x, cells[cellStart].y);
		cellStart++;
		if (cellStart == maxCells) cellStart = 0;
		int start = cellStart;
		if (start != cellEnd) {
			start++;
			if (start == maxCells) start = 0;
		}
		plotCellImage(cells[cellStart].x, cells[cellStart].y, tails[cells[start].direction]);
	}
}

int isGameOver()
{
	int lastX = cells[cellEnd].x;
	int lastY = cells[cellEnd].y;
	int gameOver = 0;
	if (lastX >= 22) gameOver = 1;
	if (lastX < 1) gameOver = 1;
	if (lastY >= 16) gameOver = 1;
	if (lastY < 1) gameOver = 1;
	int start = cellStart;
	while (start != cellEnd) {
		if (cells[start].x == lastX && cells[start].y == lastY) {
			gameOver = 1;
			break;
		}
		start++;
		if (start == maxCells) start = 0;
	}
	return gameOver;
}
	
int main(void)
{
	cells = (Cell*) malloc(maxCells * sizeof(Cell));
	initSeeds();
	SetupCallbacks();

	pgInit();
	pgScreenFrame(1, 1);
	
	newGame();
	
	while (1) {
		int i = 0;
		for (i = 0; i < 5; i++) keyboardControl();
		move();
		plotCellImage(target.x, target.y, targetImage);
		if (score > high) high = score;
		if (isGameOver()) {
			pgWaitVn(50);
			newGame();
		}
		printDecimal(410, 81, TEXT_COLOR, score);
		printDecimal(429, 129, TEXT_COLOR, high);
	}
	return 0;
}
